#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import argparse
from textwrap import dedent


class ArgparseMixin:
    """Argparse utils

    Usage
    ::
        from kipp3.options import options as opt

        # set command argument
        opt.add_argument('--test', type=str)

        # parse argument
        opt.parse_args()

        # use argument
        print(opt.test)
    """

    def setup_argparse(self, name=""):
        """Setup argparse with name"""
        self._parser = argparse.ArgumentParser(
            name, formatter_class=argparse.RawTextHelpFormatter
        )

    def add_argument(self, *args, **kw):
        """Add command arguments"""
        if not hasattr(self, "_parser"):
            self.setup_argparse()

        if "\n" in kw.get("help", ""):
            kw["help"] = dedent(kw["help"])

        self._parser.add_argument(*args, **kw)

    def parse_args(self, *args, **kw):
        """Parse command arguments"""
        is_patch_utilies = kw.pop("is_patch_utilies", True)
        args = self._parser.parse_args(*args, **kw)
        self.set_command_args(args, is_patch_utilies=is_patch_utilies)
        return args



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from threading import RLock
import time
import os
import tempfile
from functools import wraps
from collections import namedtuple
import re

try:
    import fcntl
except ImportError:
    fcntl = None

from .logger import setup_logger, get_logger
from .date import UTC, CST, parse_dtstr, utcnow, cstnow
from .mailsender import EmailSender
from .dfa_filters import DFAFilter


email_sender = EmailSender()


def run_command(command, timeout):
    """Run command and return stdout and strerr

    Args:
        command (str): command to run
        timeout (int): seconds
    """
    logger.debug("run command, {}".format(command))
    p = subprocess.Popen(
        [command], shell=True, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    try:
        outs, errs = p.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        p.kill()
        outs, errs = p.communicate()

    assert p.returncode == 0, errs
    return outs


def timer(func):
    @wraps(func)
    def wrapper(*args, **kw):
        logger.info("{} running...".format(func.__name__))
        try:
            return func(*args, **kw)
        finally:
            logger.info("{} end".format(func.__name__))

    return wrapped


class IOTA(object):
    """Simple Counter

    Examples:
    ::
        iota = IOTA()
        iota()           # return 0
        iota()           # return 1
        iota.count()     # return 2
        iota(2)          # return 4
        iota.latest()    # return 4
    """

    def __init__(self, init=-1, step=1):
        init = int(init)
        self.__count = init
        self.__step = step
        self.__lock = RLock()

    def count(self, step=None):
        with self.__lock:
            step = self.__step if step is None else int(step)
            self.__count += step
            return self.__count

    def latest(self):
        return self.__count

    def __call__(self, step=1):
        return self.count(step)

    def __str__(self):
        return str(self.__count)

    __repr__ = __str__


def sleep(secs):
    sleep_until = time.time() + secs
    remains = secs
    while remains:
        time.sleep(remains)
        remains = max(sleep_until - time.time(), 0)


VALID_FNAME_REGEX = re.compile("[a-zA-Z0-9]+")


def generate_validate_fname(val, dirpath=tempfile.gettempdir()):
    """Remove all invalidate characters in file path

    Args:
        val (str): file name
        dirpath (str, default=<system tempfile directory>):

    Returns:
        str: absolute file path
    """
    fname = "{}.lock".format("_".join(VALID_FNAME_REGEX.findall(val)))
    if dirpath:
        fname = os.path.join(dirpath, fname)

    return fname


def check_is_allow_to_running(lock_fname):
    """Check whether is another process is still running

    Args:
        lock_fname (str): the lock file path
            such as ``/mnt/log/ramjet-driver.lock``

    Returns:
        fp/False: is allow to running for current process
            fp: no other process is running, please keep the fp's reference
            False: another process is running, you should better quit current process

    Examples:
    ::
        import os

        from kipp3.utils import check_is_allow_to_running

        is_allow_running = check_is_allow_to_running('/mnt/log/ramjet-driver.lock')
        if not is_allow_running:
            print('there is another process is still running, eixt...')
            os._exit(0)
    """
    if not fcntl:
        raise NotImplementedError("not support fcntl in your system")

    fp = open(lock_fname, "w")
    try:
        fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        return False
    else:
        return fp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
------------
Email Sender
------------

Usage
::
    from kipp3.utils import EmailSender

    sender = EmailSender(opt.SMTP_HOST)  # SMTP_HOST is different between NOVA and PRD
    receivers = ','.join([xxx, xxx, xxx])

    sender.send_email(
        mail_from='data@movoto.com',
        mail_to=receivers,
        subject='Email Title',
        content='Email content'
    )

"""


from __future__ import unicode_literals
import smtplib
from textwrap import dedent
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from .logger import get_logger as get_kipp_logger


HOST = "smtp.internal.ng.movoto.net"
FROM_ADDRESS = "mls-normalizer-ng@movoto.com"


class EmailSender(object):
    def __init__(
        self,
        host=HOST,
        port=None,
        username=None,
        passwd=None,
        logger=None,
        use_tls=True,
    ):
        """Initiallize EmailSender

        Args:
            host (str, default movoto): SMTP server host
            port (int, default=None): SMTP server port
            logger (logging.logger, default=kipp_internal_logger):
            use_tls (bool, default=True)
        """
        self._host = host
        self._port = port
        self._user = username
        self._passwd = passwd
        self._logger = logger
        self._use_tls = use_tls

    def set_smtp_host(self, host):
        assert host, "smtp host should not be empty"
        assert isinstance(host, str), "smtp host should be string"
        self._host = host

    def set_smtp_port(self, port):
        assert port, "smtp port should not be empty"
        assert isinstance(port, int), "smtp port should be string"
        self._port = port

    def get_logger(self):
        return self._logger or get_kipp_logger().getChild("email")

    # def setup_by_utilities(self):
    #     try:
    #         from Utilities.movoto import settings
    #     except ImportError:
    #         return

    #     self._host = settings.HOST
    #     self._mail_from = settings.FROM_ADDRESS

    def parse_content(self, content):
        return '<font face="Microsoft YaHei, Helvetica Neue, Helvetica">{}</font>'.format(
            "<p>{}</p>".format("</p><p>".join(content.splitlines()))
        )

    def get_html(self, body):
        return """
            <head>
            <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
            <title>html title</title>
            <style type="text/css" media="screen">
                table{{
                    background-color: #f9f9f9;
                    empty-cells: hide;
                }}
            </style>
            </head>
            <body>
                {body}
            </body>
            """.format(
            body=body
        )

    def _filter_mail_to(self, mail_to):
        return ",".join(set(mail_to.split(",")))

    def send_email(
        self, mail_to, subject, content=None, html=None, mail_from=FROM_ADDRESS
    ):
        """Send email

        Args:
            mail_from (str, default to Utilities' setting): address that send from
            mail_to (str): address that send to, use `,` to split multiply receivers
            subject (str):
            content (str):

        Returns:
            bool: is successed
        """
        msg = MIMEMultipart("alternative")
        msg.set_charset("utf-8")
        if content:
            msg.attach(MIMEText(self.parse_content(content), "html"))

        if html:
            msg.attach(MIMEText(self.get_html(html), "html"))

        msg["Subject"] = subject
        msg["From"] = mail_from
        msg["To"] = self._filter_mail_to(mail_to)

        try:
            s = smtplib.SMTP(self._host, self._port)
            if self._use_tls:
                s.starttls()

            if self._user and self._passwd:
                s.login(self._user, self._passwd)

            s.sendmail(mail_from, mail_to.split(","), msg.as_string())
            s.quit()
            self.get_logger().info(
                "send email successfully to {} with subject {}".format(mail_to, subject)
            )
            return True
        except Exception:
            self.get_logger().exception(
                "fail to send email to {} with subject {} for error:".format(
                    mail_to, subject
                )
            )
            return False

    def generate_table(self, heads, contents):
        """Generate the html of table in email

        Args:
            heads (list): Table head
            contents (list of lists): Table contents

        Returns:
            str:

        Examples:
        ::
            heads = ('head 1', 'head 2', 'head 3')
            contents = (
                ('cell 1-1', 'cell 1-2', 'cell 1-3'),
                ('cell 2-1', 'cell 2-2', 'cell 2-3')
            )

            table_html = sender.generate_table(heads, contents)
        """
        thead = "".join(["<th><p>{}</p></th>\n".format(str(h)) for h in heads])
        tbody = ""
        for cnt in contents:
            tbody += "<tr>{}</tr>\n".format(
                "".join(["<td><p>{}</p></td>".format(str(h)) for h in cnt])
            )

        return dedent(
            """
            <table style="table-layout:fixed;" cellspacing="0" cellpadding="10">
                <thead><tr>{thead}</tr></thead>
                <tbody>{tbody}</tbody>
            </table>
            """
        ).format(thead=thead, tbody=tbody)


if __name__ == "__main__":
    sender = EmailSender()
    assert sender.send_email(
        mail_to="lcai@movoto.com,lcai@movoto.com",
        subject="test: kipp.utils.EmailSender",
        content="fake-content\n\nline 2",
        mail_from="kipp@movoto.com",
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
---------------
Datatime Utilis
---------------

Usage
::

    # import timezone constants
    from kipp3.utils import UTC, CST

    # import datetime functions
    from kipp3.utils import parse_dtstr, utcnow, cstnow

"""

from __future__ import unicode_literals
import re
import datetime
from future.utils import iteritems

import pytz
import maya


UTC = pytz.timezone("utc")
CST = pytz.timezone("Asia/Shanghai")

_SPECIAL_DTSTR_REGEX_MAP = {
    re.compile(r"^(\d{4})$"): "%H%M",
    re.compile(r"^(\d{2}:\d{2}) noon$", flags=re.I): "%H:%M",
}


def _parse_special_dtstr(dtstr):
    """Parse MLSes weird datetime string formats"""
    for dt_regx, dt_fmt in iteritems(_SPECIAL_DTSTR_REGEX_MAP):
        m = dt_regx.match(dtstr)
        if m:
            _dtstr = m.groups()[0]
            return UTC.localize(datetime.datetime.strptime(_dtstr, dt_fmt))


_IGNORE_TZ_REGEX = re.compile(r"[+\-][0-9:]{1,5}$")


def _extrace_dtstr_exclude_tz(dtstr):
    r = _IGNORE_TZ_REGEX.search(dtstr)
    if not r:
        return dtstr

    return dtstr[: r.start()]


def parse_dtstr(
    date_str, naive=False, replace_tz=None, convert_tz=None, ignore_tz=False
):
    """Parse datetime string to datetime object

    Default timezone is UTC.

    Args:
        date_str (str): origin datetime string.
        naive (bool, default=False): if True, return a naive datetime object.
        replace_tz (timezone):replace to che specified timezone,
            without change the literal datetime attributes.
        convert_tz (timezone, default=UTC): convert to specified timezone.
        ignore_tz (bool, default=False): do not parse timezone.

    Raises:
        ValueError: if cannot parse the date_str.

    Examples:
    ::
        from kipp3.utils import parse_dtstr

        dtstr = '2017/1/1 13:00 +08:00'
        dt = parse_dtstr(dtstr)

    """
    if ignore_tz:
        date_str = _extrace_dtstr_exclude_tz(date_str)

    dt = _parse_special_dtstr(date_str) or maya.parse(date_str).datetime()
    if naive:
        dt = dt.replace(tzinfo=None)
    elif replace_tz:
        dt = dt.replace(tzinfo=replace_tz)
    elif convert_tz:
        dt = dt.astimezone(tz=convert_tz)

    return dt


def utcnow(is_naive=False):
    """Get current datetime with UTC timezone"""
    dt = datetime.datetime.utcnow()
    if not is_naive:
        dt = UTC.localize(dt)

    return dt


def cstnow(is_naive=False):
    """Get current datetime with cst timezone"""
    dt = utcnow()
    if not is_naive:
        dt = dt.astimezone(CST)

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import functools
import traceback

try:
    from time import monotonic as time
except ImportError:
    from time import time
import os
import sys
import datetime
import signal
from collections import namedtuple, defaultdict

import xxhash

from kipp3.utils import get_logger


def retry(ExceptionToCheck, tries=3, delay=1, backoff=1):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: int
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: int
    :param logger: logger to use. If None, print
    :type logger: logging.Logger instance
    """

    def deco_retry(f):
        @functools.wraps(f)
        def _wrapper(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck:
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff

            return f(*args, **kwargs)

        return _wrapper  # true decorator

    return deco_retry


def single_instance(pidfilename, logger=None):
    def create_pid(pidfilename):
        current_pid = os.getpid()
        pidfile = open(pidfilename, "w")
        pidfile.write(str(current_pid))
        pidfile.close()

    def read_pid(file_path):
        """Read pid from a pid file"""
        f = open(file_path)
        pidv = f.read()
        return pidv.strip()

    def check_pid(pid):
        """ Check For the existence of a unix pid. """
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def deco_single_instance(func):
        """Make sure only one instance of this program with the same parameters runs"""
        if os.path.exists(pidfilename):
            pidv = read_pid(pidfilename)
            if check_pid(int(pidv)):
                get_logger()().info(
                    "There's already an instance of this program, pid : %s", pidv
                )
                sys.exit()
            else:
                os.remove(pidfilename)
        create_pid(pidfilename)

        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return _wrapper  # true decorator

    return deco_single_instance


def timer(fn):
    """decorator to log running time

    Examples:
    ::
        from kipp3.decorator import timer

        @timer
        def demo():
            time.sleep(10)

    """
    @functools.wraps(fn)
    def wrapper(*args, **kw):
        try:
            start_at = time()
            r = fn(*args, **kw)
        except Exception as err:
            get_logger().exception("run {}".format(fn.__name__))
            raise
        else:
            return r
        finally:
            get_logger().info("{} cost {:.2f}s".format(fn.__name__, time() - start_at))

    return wrapper


debug_wrapper = timer # compatable


def memo(fn):
    cache = {}
    miss = object()

    @functools.wraps(fn)
    def wrapper(*args):
        result = cache.get(args, miss)
        if result is miss:
            result = fn(*args)
            cache[args] = result
        return result

    return wrapper


class TimeoutError(Exception):
    def __init__(self, value="Timed Out"):
        self.value = value

    def __str__(self):
        return repr(self.value)


def timeout(seconds_before_timeout):
    def decorate(f):
        def handler(signum, frame):
            raise TimeoutError()

        def new_f(*args, **kwargs):
            old = signal.signal(signal.SIGALRM, handler)
            signal.alarm(seconds_before_timeout)
            try:
                result = f(*args, **kwargs)
            finally:
                signal.signal(signal.SIGALRM, old)
            signal.alarm(0)
            return result

        new_f.func_name = f.func_name
        return new_f

    return decorate


def calculate_args_hash(*args, **kw):
    return xxhash.xxh32(str(args) + str(kw)).hexdigest()


CacheItem = namedtuple("item", ["data", "timeout_at"])


def timeout_cache(expires_sec=30, max_size=128):
    """Decorator to cache return until expires

    Args:
        expires_sec (int): cache expires time

    Examples:
    ::
    import time

    from kipp3.utils import timeout_cache

    @timeout_cache(expires_sec=2)
    def demo():
        return time.time()

    r = demo()
    time.sleep(1)
    r == demo()

    """
    assert expires_sec > 0, "expires_sec should greater than 0, but got {}".format(
        expires_sec
    )
    assert max_size > 0, "max_size should greater than 0, but got {}".format(max_size)
    state = {}  # {hkey: CacheItem}

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kw):
            hkey = calculate_args_hash(*args, **kw)
            if hkey not in state or state[hkey].timeout_at < time():
                if len(state) > max_size:  # remove expired keys
                    for k in list(state.keys()):
                        if state[k].timeout_at > time():
                            del state[k]

                state[hkey] = CacheItem(
                    timeout_at=time() + expires_sec, data=f(*args, **kw)
                )

            return state[hkey].data

        return wrapper

    return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import


class KippException(Exception):
    pass


class KippAIOException(KippException):
    pass


class KippAIOTimeoutError(KippAIOException):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
----------
Kipp Utils
----------


"""

from __future__ import unicode_literals
import sys
import time
import logging


LOGNAME = "kipp"  # kipp internal logger name


def get_formatter():
    formatter = logging.Formatter(
        "[%(asctime)s - %(levelname)s - %(pathname)s:%(lineno)d - %(name)s] - %(message)s"
    )
    formatter.converter = time.gmtime
    return formatter


def get_stream_handler():
    formatter = get_formatter()
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter)
    return ch


def setup_logger(logname, debug=False):
    """Create logger with default formatter and stream of stdout

    Args:
        logname (str): the name of the logger
        debug (bool, default=False): whether set logging level to DEBUG
    """
    logger = logging.getLogger(logname)
    log_level = logging.DEBUG if debug else logging.INFO

    # logger.setLevel(logging.DEBUG)
    # logger.setLevel(log_level)
    stream_handler = get_stream_handler()
    logging.getLogger("tornado").addHandler(stream_handler)
    logging.getLogger("tornado").setLevel(logging.ERROR)
    logging.getLogger("concurrent").addHandler(stream_handler)
    logging.getLogger("concurrent").setLevel(logging.ERROR)

    # root_logger = logging.getLogger()
    logger.setLevel(log_level)
    if not logger.handlers:
        logger.addHandler(stream_handler)

    return logger


_logger = {"ins": setup_logger(logname=LOGNAME)}  # do not change directly


def get_logger():
    """Get kipp internal logger"""
    return _logger["ins"]


def get_wrap_handler(target_logger):
    class _WrapperHandler(logging.StreamHandler):
        def emit(self, record):
            formatter = logging.Formatter("[%(levelname)s:%(name)s] %(message)s")
            target_logger.log(record.levelno, formatter.format(record))

    return _WrapperHandler()


def set_logger(logger):
    """Replace kipp internal logger

    Since of only Utilities' logger can output to file,
    so you need to replace the kipp' internal logger with Utilities' logger
    to save your logs to file.

    Usage:
    ::
        set_logger(utilities_logger)
    """
    handler = get_wrap_handler(logger)
    get_logger().addHandler(handler)
    for dep_logger_name in ("tornado", "concurrent"):
        logging.getLogger(dep_logger_name).addHandler(handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals


class DFAFilter:
    """
    Examples:
    ::
        dfa_filter = DFAFilter()
        dfa_filter.build_chains(keywords_set)
        dfa_filter.load_keywords(raw_text)
    """

    def load_keywords(self, raw_text):
        """
        Args:
            raw_text (str):

        Returns:
            set: keywords that in raw_text
        """
        assert getattr(self, "_chains", None), "Should invoke build_chains first"
        return self.filter_keyword(raw_text)

    def build_chains(self, keywords):
        """
        Args:
            keywords (set): lexicon of keywords
        """
        chains = {}
        for word in keywords:
            node = chains
            for char in word:
                if char not in node:
                    node[char] = {}

                node = node[char]

        self._chains = chains

    def is_word_in_chains(self, chains, raw_text, n_len, i):
        if raw_text[i] not in chains:
            return None

        if not chains[raw_text[i]]:
            return i

        if i == n_len - 1:
            return None

        return self.is_word_in_chains(
            chains=chains[raw_text[i]], raw_text=raw_text, n_len=n_len, i=i + 1
        )

    def filter_keyword(self, raw_text):
        result_keywords = set()
        i, n_len = 0, len(raw_text)
        for i in range(n_len):
            li = self.is_word_in_chains(self._chains, raw_text, n_len, i)
            if li is not None:
                result_keywords.add(raw_text[i : li + 1])

        return result_keywords



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
-------------------------
Image resize and compress
-------------------------

resize and compress images with width and quality.

Notify: You should run ``pip install "kipp[image]"`` to install requirements.

"""

from __future__ import unicode_literals
from io import BytesIO

from PIL import Image

from kipp3.decorator import retry


IMG_CONTENT_TYPE = "image/jpeg"
THUMBNAIL_TYPES = (("p", 480), ("r", 960), ("l", 1920))


def image_resize_and_compress(image_bytesIO, width, quality=75, optimize=True):
    """Image resize and compress

    Relate to 'DATA-1205'.

    Args:
        image_data (file handler): origin image.
        width (int): if origin image is wider than width, resize to width.
        quality (int, default=80): quality of destination image.
        optimize (bool, default=True): if True, return optimized image.
    """
    pil_image = Image.open(image_bytesIO)
    if pil_image.size[0] <= width:
        return compress_image(image_bytesIO, quality, optimize)

    return resize_compress_image(image_bytesIO, width, quality, optimize)


def compress_image(image_bytesIO, quality=75, optimize=True):
    pil_image = Image.open(image_bytesIO)
    img_data = BytesIO()
    pil_image.convert("RGB").save(
        img_data, format="JPEG", quality=quality, optimize=optimize
    )
    img_data.seek(0)
    image_bytesIO.seek(0)
    return (
        image_bytesIO if len(img_data.read()) > len(image_bytesIO.read()) else img_data
    )


def resize_compress_image(image_bytesIO, width, quality=75, optimize=True):
    """Get the thumbnail image with input bytesIO and thumbnail width

    This function is deprecated, because new thumbnail image bytes may be more than original image.
    """
    pil_image = Image.open(image_bytesIO)
    (x, y) = pil_image.size
    new_size = (width, int(width * y / x))
    resized_img_bytesio = BytesIO()
    pil_image.resize(new_size, Image.ANTIALIAS).convert("RGB").save(
        resized_img_bytesio, format="JPEG", quality=quality, optimize=optimize
    )
    return resized_img_bytesio


def get_thumbnail(
    input_image_bytesIO,
    thumbnail_width,
    thumbnail_height=0,
    quality_withdraw=5,
    quality=75,
    optimize=True,
):
    """Generate the image thumbnails for one input image

    Args:
        input_image_bytesIO (io.BytesIO): input origin image bytesIO.
        thumbnail_width (int): thumbnail image width needed
        quality_withdraw (int): quality_withdraw
        quality & optimize: see Pillow document.
    """
    input_image_bytesIO.seek(0)
    input_bytes = len(input_image_bytesIO.read())
    pimg = Image.open(input_image_bytesIO)

    if not thumbnail_height:
        thumbnail_height = int(thumbnail_width * pimg.size[1] / pimg.size[0])
    thm_size = (thumbnail_width, thumbnail_height)
    pimg = pimg.convert("RGB").resize(thm_size, Image.ANTIALIAS)
    q = quality
    while 1:
        thumbnail_data = BytesIO()
        pimg.save(thumbnail_data, format="JPEG", quality=q, optimize=optimize)
        thumbnail_data.seek(0)
        t_bytes = len(thumbnail_data.read())

        if t_bytes > input_bytes:
            q -= quality_withdraw
        else:
            thumbnail_data.seek(0)
            return thumbnail_data


def generate_thumbnails(image_bytesIO, img_s3_key, s3):
    """Generate the image thumbnails for one input image,
        and store the thumbnails in s3.

    Args:
        img_data (io.BytesIO): origin image.
        img_s3_key (str): img_s3_key
        s3 (Utilities.movoto.s3_handler.S3_handler): s3 connected client

    """
    pil_image = Image.open(image_bytesIO)
    key_slices = img_s3_key.split(".")
    copy_key = ""
    for tname, twidth in THUMBNAIL_TYPES:
        thumnail_s3_key = "{}_{}.{}".format(key_slices[0], tname, key_slices[1])
        if copy_key:
            s3.copy_file(copy_key, thumnail_s3_key)
            continue

        if pil_image.size[0] <= twidth:
            thm_data = compress_image(image_bytesIO)  # only do compress
            copy_key = thumnail_s3_key
        else:
            thm_data = get_thumbnail(image_bytesIO, twidth)
        thm_data.seek(0)
        resp = s3.m_uploadImage(thm_data, IMG_CONTENT_TYPE, thumnail_s3_key)
        if resp.status != 200:
            raise IOError(
                "UploadS3Error!! Key[%s] Response[status: %s, reason: %s, msg: %s]"
                % (thumnail_s3_key, resp.status, resp.reason, resp.msg)
            )

    flag = 2 ** len(THUMBNAIL_TYPES) - 1
    return flag


# def generate_thumbnails(image_bytesIO, img_s3_key, s3):
#     """ Generate the image thumbnails for one input image,
#         and store the thumbnails in s3.
#
#     Args:
#         img_data (io.BytesIO): origin image.
#         img_s3_key (str): img_s3_key
#         thumbnail_types : the thumbnails to generate, please make sure it is order by width ascend.
#                         Examples: ((u'p', 480),
#                                     (u'r', 960),
#                                     (u'l', 1920))
#
#         s3 (Utilities.movoto.s3_handler.S3_handler): s3 connected client
#
#     """
#     pil_image = Image.open(image_bytesIO)
#     key_slices = img_s3_key.split(".")
#     copy_key = ''
#     thum_o = None
#     for tname, twidth in THUMBNAIL_TYPES[::-1]:
#         thumnail_s3_key = "{}_{}.{}".format(key_slices[0], tname, key_slices[1])
#         if pil_image.size[0] <= twidth:
#             if copy_key:
#                 new_key = s3.copy_file(copy_key, thumnail_s3_key)
#                 if not new_key:
#                     raise Exception('Copy S3 key [%s] error!' % thumnail_s3_key)
#                 continue
#             else:
#                 thm_data = compress_image(image_bytesIO) # only do compress
#                 thum_o = thm_data
#                 copy_key = thumnail_s3_key
#         else:
#             thm_data = get_thumbnail(thum_o if thum_o else image_bytesIO, twidth)
#         thm_data.seek(0)
#         resp = s3.m_uploadImage(thm_data, IMG_CONTENT_TYPE, thumnail_s3_key)
#         if resp.status != 200:
#             raise Exception('Upload S3 key [%s] error!' % thumnail_s3_key)
#
#     flag = 2 ** len(THUMBNAIL_TYPES) - 1
#     return flag


def generate_possible_thumbnails(image_bytesIO, img_s3_key, s3):
    """Generate the image thumbnails for one input image,
        and store the thumbnails in s3.

    Args:
        img_data (io.BytesIO): origin image.
        img_s3_key (str): img_s3_key
        s3 (Utilities.movoto.s3_handler.S3_handler): s3 connected client

    """
    pil_image = Image.open(image_bytesIO)
    key_slices = img_s3_key.split(".")
    generated_thumbnail_list = []
    for tname, twidth in THUMBNAIL_TYPES:
        if pil_image.size[0] > twidth:
            thumnail_s3_key = "{}_{}.{}".format(key_slices[0], tname, key_slices[1])
            theight = int(twidth * pil_image.size[1] / pil_image.size[0])
            thm_data = get_thumbnail(image_bytesIO, twidth, theight)
            thm_data.seek(0)
            resp = s3.m_uploadImage(thm_data, IMG_CONTENT_TYPE, thumnail_s3_key)
            if resp.status != 200:
                raise IOError(
                    "UploadS3Error!! Key[%s] Response[status: %s, reason: %s, msg: %s]"
                    % (thumnail_s3_key, resp.status, resp.reason, resp.msg)
                )
            generated_thumbnail_list.append(
                {"width": twidth, "type": tname, "height": theight}
            )
        else:
            break
    return generated_thumbnail_list


@retry(Exception, tries=3, delay=3, backoff=3)
def request_s3_image(url, http_session):
    """Request image data by http session

    Args:
        url (str): request url
        http_session: you can get it yourself by requests.Session() or give requests.
    """
    res = http_session.get(url)
    assert res.status_code == 200, "request_s3_image error"
    bytesIO = BytesIO()
    bytesIO.write(res.content)
    bytesIO.seek(0)
    return bytesIO


if __name__ == "__main__":
    import requests as http

    # url = "http://staging-img.movoto.com.s3.amazonaws.com/p/404/9548575_0_nyyVEa.jpeg"
    key = "p/104/477149_0_iQyiAu.jpeg"
    #     origin_image_data = request_s3_image("http://staging-img.movoto.com.s3.amazonaws.com/"+key, http)
    #     with open("/tmp/"+key, 'w') as fobj:
    #         fobj.write(origin_image_data.read())
    #     origin_image_data.seek(0)
    import os

    origin_image_data = open("/tmp/" + key, "rb")

    img = Image.open(origin_image_data)
    print(img.size)
    origin_image_data.seek(0)
    import time

    t0 = time.time()
    thumbnail_img_data = get_thumbnail(origin_image_data, 960)
    print(time.time() - t0)
    save_path = "/tmp/p/104/477149_0_iQyiAu_960.jpeg"
    with open(save_path, "w") as fobj:
        fobj.write(thumbnail_img_data.read())

    print("/tmp/" + key, os.path.getsize("/tmp/" + key))
    print(save_path, os.path.getsize(save_path))
    quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
---------------------
Configuration Manager
---------------------

Uniform arguments entrypoint.

"""

from __future__ import unicode_literals, absolute_import
import os
import copy

# import sys
import importlib

# from imp import load_source

from kipp3.libs import SingletonMixin
from kipp3.utils import get_logger

# from .mlogger import LazyMLogger, ConvertFailureLog
from .arguments import ArgparseMixin
from .exceptions import KippOptionsException


# class UtilitiesMixin:
#     """Load Utilities from PYTHONPATH
#     """

#     def load_utilities_settings(self):
#         for p in sys.path:
#             py_path = os.path.abspath(p)
#             if not os.path.isdir(py_path):
#                 continue

#             for package in next(os.walk(py_path))[1]:
#                 if package != 'Utilities':
#                     continue

#                 if not os.path.isdir(os.path.join(p, package)):
#                     continue

#                 st_fpath = os.path.join(p, package, 'movoto', 'settings.py')
#                 get_logger().info('load utilities from: %s', st_fpath)
#                 return load_source('utils_settings', st_fpath)

#         raise KippOptionsException('Can not found ``Utilities`` in PYTHONPATH')

#     def patch_utilities(self):
#         """Patch Utilities settings"""
#         if getattr(self, '_is_patched_utilities', None):  # do not patch twice
#             return

#         import Utilities
#         from Utilities import movoto as utils_movoto
#         from Utilities.movoto import settings as utils_settings

#         self._orig_utilities = Utilities
#         self._orig_movoto = utils_movoto
#         self._movoto_settings = utils_settings

#         _self = self

#         class SettingsMock(object):

#             def __getattr__(self, name):
#                 if getattr(_self, '_is_patched_utilities', None):
#                     return getattr(_self, name)
#                 else:
#                     return getattr(_self._movoto_settings, name)

#         class MovotoMock(object):

#             def __getattr__(self, name):
#                 if name == 'settings':
#                     return SettingsMock()
#                 else:
#                     return getattr(_self._orig_movoto, name)

#         class UtilitiesMock(object):

#             def __getattr__(self, name):
#                 if name == 'movoto':
#                     return MovotoMock()
#                 else:
#                     return getattr(_self._orig_utilities, name)

#         sys.modules['Utilities'] = UtilitiesMock()
#         sys.modules['Utilities.movoto'] = MovotoMock()
#         sys.modules['Utilities.movoto.settings'] = SettingsMock()

#         self._is_patched_utilities = True

#     def unpatch_utilities(self):
#         if not getattr(self, '_is_patched_utilities', None):
#             return

#         sys.modules['Utilities'] = self._orig_utilities
#         sys.modules['Utilities.movoto'] = self._orig_movoto
#         sys.modules['Utilities.movoto.settings'] = self._orig_movoto.settings
#         self._is_patched_utilities = False


# class MLoggerMixin:
#     def get_mlogger(self, *args, **kw):
#         """(deprecated) Get lazy-init mlogger

#         If pass no arguments, will return last setuped logger

#         DEPRECATED:
#             now you can set logger, then save logger in opt by ``opt.set_option('logger', logger)``
#         """
#         if getattr(self, '_mlogger', None) and not args and not kw:
#             return self._mlogger
#         else:
#             self._mlogger = LazyMLogger(*args, **kw)
#             return self._mlogger


class OptionKeyTypeConflictError(KippOptionsException):
    pass


class BaseOptions(object):
    def __init__(self, is_allow_overwrite_child_opt=True):
        self._inner_settings = {}
        self._is_allow_overwrite_child_opt = is_allow_overwrite_child_opt

    def create_option(self):
        return BaseOptions()

    @property
    def is_allow_overwrite_child_opt(self):
        return self._is_allow_overwrite_child_opt

    @is_allow_overwrite_child_opt.setter
    def is_allow_overwrite_child_opt(self, val):
        assert isinstance(val, bool), "must be bool"
        self._is_allow_overwrite_child_opt = val

    def set_option(self, name, val):
        """Set your own attribute's name and value

        If name contains `.`, will auto create child option
        """
        if "." not in name:
            self._inner_settings[name] = val
            return

        children = name.split(".")
        opt = self
        for child in children[:-1]:
            if child not in opt:
                opt.set_option(child, self.create_option())
            elif not isinstance(opt[child], BaseOptions):
                if self._is_allow_overwrite_child_opt:
                    opt.set_option(child, self.create_option())  # overwrite
                else:
                    raise OptionKeyTypeConflictError(
                        "{} already exists but type is not BaseOption".format(child)
                    )

            opt = opt[child]

        last_key = children[-1]
        if (
            last_key in opt
            and isinstance(opt[last_key], BaseOptions)
            and not self._is_allow_overwrite_child_opt
        ):
            raise OptionKeyTypeConflictError("must not overwrite child option")

        opt.set_option(last_key, val)

    __setitem__ = set_option

    def del_option(self, name):
        """Delete your own attribute by name"""
        if name in self._inner_settings:
            del self._inner_settings[name]

    __delitem__ = del_option

    def get_option(self, name):
        """If you overwrite `get_option`, you need rewrite
        `__getitem__ = __getattr__ = get_option`
        """
        if "." not in name:
            if name not in self._inner_settings:
                raise AttributeError
            else:
                return self._inner_settings[name]

        opt = self
        for child in name.split("."):
            if not isinstance(opt, BaseOptions) or child not in opt:
                raise AttributeError

            opt = opt[child]

        return opt

    __getitem__ = __getattr__ = get_option

    def __contains__(self, name):
        """Override ``in`` operator"""
        try:
            self.get_option(name)
        except AttributeError:
            return False
        else:
            return True


class Options(BaseOptions, ArgparseMixin, SingletonMixin):
    """Configuration Manager

    Relate to `DATA-1026`_.

    You can write your own ``settings/settings_xxx.py`` to overwrite the Utilities settings.
    Please double check ``settings/__init__.py`` is exists.

    Load arguments in following sequences:

    #. load from your own attributes;
    #. load from command arguments (argparse);
    #. load from OS environment;
    #. load from env specified settings (``settings_xxx.py``);
    #. load from ``settings/settings_local.py``;
    #. load from project settings (``settings/settings.py``);
    #. load from ``Utilities.movoto.settings``;


    Usage:

        General usage::

            from kipp3.options import options as opt

            # both ok
            opt.get_option('debug')
            opt.debug
            opt['debug']

            # will raise AttributeError if not exists

        Setup with argparse::

            from kipp3.options import options as opt

            # get args from argparse
            parser = argparse.ArgumentParser()
            args = parser.parse_args()

            # set args
            opt.set_command_args(args)

            # if you want patch utilities at the same time
            opt.set_command_args(args, is_patch_utilies=True)

        Patch Utilities only::

            from kipp3.options import options as opt

            # should patch before import Utilities
            opt.patch_utilities()

            from Utilities.movoto import settings

        Setup your own attribute::

            from kipp3.options import options as opt

            # 321
            # assume init value is 321

            # you can set your own attr value to any object
            opt.set_option('abc', 123)
            # 123

            # then delete it to restore the default value
            opt.del_option('abc')
            # 321


    .. _DATA-1026:
        https://movoto.atlassian.net/browse/DATA-1026

    """

    def __init__(self, is_allow_overwrite_child_opt=True):
        super(Options, self).__init__(
            is_allow_overwrite_child_opt=is_allow_overwrite_child_opt
        )
        self.setup_env_settings()

    def setup_env_settings(self):
        self._command_args = None
        self._environ = copy.deepcopy(os.environ)
        self._env_settings = None
        self._private_settings = None

        # try:  # load Utilities settings
        #     movoto_settings = self.load_utilities_settings()
        # except KippOptionsException:
        #     self._movoto_settings = None
        #     get_logger().warning('can not found ``Utilities`` in PYTHONPATH')
        # else:
        #     get_logger().info('setup settings from Utilities')
        #     self._movoto_settings = movoto_settings

        try:  # load private settings
            from settings import settings_local as private_settings
        except ImportError:
            self._private_settings = None
        else:
            get_logger().info("setup settings from settings_local.py")
            self._private_settings = private_settings

        try:  # load project settings
            from settings import settings as project_settings
        except ImportError:
            self._project_settings = None
        else:
            get_logger().info("setup settings from settings.py")
            self._project_settings = project_settings

        self.load_specifical_settings()  # load env from environment

    def get_option(self, name):
        try:
            val = super(Options, self).get_option(name)
        except AttributeError:
            pass
        else:
            get_logger().debug("load attr %s from inner settings", name)
            return val

        if hasattr(self._command_args, name):
            get_logger().debug("load attr %s from command args", name)
            return getattr(self._command_args, name)
        elif name in self._environ:
            get_logger().debug("load attr %s from environ", name)
            return self._environ[name]
        elif hasattr(self._env_settings, name):
            get_logger().debug("load attr %s from env settings", name)
            return getattr(self._env_settings, name)
        elif hasattr(self._private_settings, name):
            get_logger().debug("load attr %s from private settings", name)
            return getattr(self._private_settings, name)
        elif hasattr(self._project_settings, name):
            get_logger().debug("load attr %s from project settings", name)
            return getattr(self._project_settings, name)
        # elif hasattr(self._movoto_settings, name):
        #     get_logger().debug('load attr from %s movoto Utilities', name)
        #     return getattr(self._movoto_settings, name)

        raise AttributeError("attribute `{}` Not Found in kipp.options".format(name))

    __getitem__ = __getattr__ = get_option

    def load_specifical_settings(self, env=None):
        env = env or os.environ.get("TARS_ENV", None)
        if not env:
            return

        try:
            settings_fname = "settings.settings_{}".format(env)
            env_settings = importlib.import_module(settings_fname)
        except ImportError:
            raise KippOptionsException("Can not found env settings_{}".format(env))
        else:
            self._env_settings = env_settings
            get_logger().info("Setup specifical settings from settings_%s", env)

    def set_command_args(self, args, is_patch_utilies=True):
        """"Setup with argparse"""
        self._command_args = args
        if hasattr(args, "env"):
            get_logger().warning(
                "You should better to use TARS_ENV environment to set run env"
            )
            self.load_specifical_settings(args.env)

        # no need for public use
        # if is_patch_utilies and not getattr(self, '_is_patched_utilities', None):
        #     self.patch_utilities()


opt = options = Options()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import
from datetime import datetime


class LazyMLogger:
    """Lazy Movoto Logger

    Load logger just before use.
    Let Utilities patch works well.

    Initialize
    ::

        from kipp3.options import options as opt

        logger = opt.get_mlogger()  # get logger without init

        logger.setup(*args, **kw)  # setup logger after opt.set_command_args

        logger.info(xxx)  # will construct MLogger just before use
    """

    _mlogger = None
    _args = _kw = None

    def __init__(self, *args, **kw):
        self._args = args
        self._kw = kw

    def __getattr__(self, name):
        if not self._mlogger:
            self.setup()

        return getattr(self._mlogger, name)

    def setup(self):
        """Setup movoto's MLogger manually"""
        from Utilities.movoto.logger import MLogger

        self._mlogger = MLogger().getLogger(*self._args, **self._kw)


class ConvertFailureLog:
    """MLS mapping failure logs

    Accroding to DATA-1531

    Usage:
    ::
        opt.setup_failure_logger(stage, mls_id)
        opt.record_failure_log(mls_sysid, mls_number, column_name, column_value, exc_info)
        # will record log to ``log_null_values_{stage}_{mls_id}.log``
    """

    def setup_failure_logger(self, stage, mls_id):
        """Setup failures logger

        Args:
            stage (str): normalizer/converter/keywords
            mls_id (int):

        Returns:
            logger: the failure logger
        """
        assert stage in (
            "normalizer",
            "converter",
            "keywords",
        ), "stage should be 'normalizer'/'converter'/'keywords'"

        from Utilities.movoto.logger import MLogger

        self._failures_logger = MLogger().getLogger(
            "log_null_values_{stage}_{mls_id}".format(stage=stage, mls_id=mls_id),
            mls_id,
        )
        return self._failures_logger

    def format_failure_log(
        self, mls_sysid, mls_number, column_name, column_value, exc_info=None
    ):
        """Format arguments to log message

        Args:
            mls_sysid (str):
            mls_number (str):
            column_name (str):
            column_value (str):
            exc_info (Exception):

        Returns:
            str: message
        """
        s = "mls_sysid: {mls_sysid}, mls_number: {mls_number}, column_name: {column_name}, column_value: {column_value}".format(
            mls_sysid=mls_sysid,
            mls_number=mls_number,
            column_name=column_name,
            column_value=column_value,
        )
        if exc_info:
            s += ", error: ".format(exc_info)

        return s

    def record_failure_log(self, **kw):
        """Record message to failure logger

        arguments as same as format_failure_log
        """
        assert (
            self._failures_logger
        ), "You should invoke ``setup_failure_logger(stage, mls_id)`` first"

        if kw.get("exc_info", None):
            self._failures_logger.error(
                self.format_failure_log(**kw), exc_info=kw["exc_info"]
            )
        else:
            self._failures_logger.info(self.format_failure_log(**kw))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from kipp3.utils import utcnow, get_logger


RUNSTATSMONITOR_COLLECTION_NAME = "run_stats_monitor"


class RunStats:
    running = "running"
    failed = "failed"
    successed = "successed"


class RunStatsMonitor(object):
    def __init__(self, command, args=None):
        self.connect()
        self._command = command
        self._args = args

    def connect(self):
        from Utilities.movoto.mongodbHelper import MongodbHelper

        self.mongo = MongodbHelper().connect()
        self.db = self.mongo.db
        self.collection = self.db[RUNSTATSMONITOR_COLLECTION_NAME]

    def start(self):
        d = {
            "command": self._command,
            "updated_at": utcnow(),
            "created_at": utcnow(),
            "status": RunStats.running,
        }
        if self._args:
            d.update({"arguments": self._args})

        r = self.mongo.create_documents(RUNSTATSMONITOR_COLLECTION_NAME, [d])
        self.task_id = r[0]
        return self.task_id

    def success(self):
        self.mongo.update_fields(
            RUNSTATSMONITOR_COLLECTION_NAME,
            {"_id": self.task_id},
            {"updated_at": utcnow(), "status": RunStats.successed},
        )

    def clean_logs_by_timedelta(self, timedelta):
        get_logger().info("clean_logs_by_timedelta for timedelta %s", timedelta)
        self.collection.remove({"updated_at": {"$lt": utcnow() + timedelta}})

    def fail(self, err_msg):
        self.mongo.update_fields(
            RUNSTATSMONITOR_COLLECTION_NAME,
            {"_id": self.task_id},
            {"updated_at": utcnow(), "status": RunStats.failed, "err_msg": err_msg},
        )

    def __del__(self):
        self.mongo.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
---------------
Kipp Exceptions
---------------

KippException
  ├─ KippAIOException
  │    └─ KippAIOTimeoutError
  ├─ DBError
  │    ├─ DBValidateError
  │    ├─ DuplicateIndexError
  │    └─ RecordNotFound
  ├─ KippRunnerException
  │    └─ KippRunnerTimeoutException

"""


from __future__ import unicode_literals

from kipp3.libs.exceptions import KippAIOException, KippAIOTimeoutError, KippException
from kipp3.options import KippOptionsException, OptionKeyTypeConflictError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import
from functools import wraps
from sys import version_info

from .exceptions import KippException, KippAIOException


PY2 = version_info[0] == 2
PY3 = version_info[0] == 3


def singleton(cls, *args, **kw):
    instances = {}

    @wraps(cls)
    def _singleton():
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton


class SingletonMixin:
    def __new__(cls, *args, **kw):
        if not hasattr(cls, "_instance"):
            orig = super(SingletonMixin, cls)
            cls._instance = orig.__new__(cls, *args, **kw)
        return cls._instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
import codecs

try:
    import setuptools
    from setuptools import setup
except ImportError:
    setuptools = None
    from distutils.core import setup

import kipp3


with open("requirements.txt") as f:
    install_requires = f.read().strip().split("\n")


def update_readme_version(version):
    ver_reg = re.compile(
        "(https://img\.shields\.io/badge/version-v"
        "[0-9]+\.[0-9]+(\.[0-9]+)?((dev|rc)[0-9]+)?"
        "-blue\.svg)"
    )
    _v = "https://img.shields.io/badge/version-v{}-blue.svg".format(version)
    with open("README.md", "r") as f:
        src = f.read()

    with open("README.md", "w") as f:
        dest = ver_reg.sub(_v, src)
        f.write(dest)


version = kipp3.__version__
update_readme_version(version)

with codecs.open("README.md", "r", "utf8") as f:
    long_description = f.read()


name = "kipp3"
packages = []
package_dir = {name: name}
for dirname, dirnames, filenames in os.walk(name):
    if "__init__.py" in filenames:
        packages.append(dirname.replace("/", "."))

extras = {
    "test": ["pytest", "mock"],
    "doc": ["sphinx", "recommonmark", "sphinxcontrib-napoleon"],
    "image": ["pillow==3.4.2"],
}
all_extras = []
for _, v in extras.items():
    all_extras.extend(v)

extras["all"] = all_extras

data_files = [
    "requirements.txt",
    "README.md",
    "CHANGELOG.md",
    "LICENSE",
    "Makefile",
    "package.json",
    "tox.ini",
]

setup(
    name=name,
    version=version,
    packages=packages,
    package_dir=package_dir,
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras,
    data_files=data_files,
    author="Laisky",
    author_email="ppcelery@gmail.com",
    description="Python Utils",
    long_description=long_description,
    url="https://github.com/Laisky/kipp",
    license="MIT License",
    entry_points="""\
        [console_scripts]
        kipp3_runner=kipp3.runner.__main__:main
    """,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Libraries",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=["setup", "distutils", "utils"],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from kipp3.libs import KippException


class KippRunnerException(KippException):
    pass


class KippRunnerTimeoutException(KippRunnerException):
    pass


class KippRunnerSIGTERMException(KippRunnerException):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
-----------
Kipp Runner
-----------

Collect scripts' running status into MongoDB

Arguments:
    -t, --timeout (int): seconds to throw ``KippRunnerTimeoutException``
    -l, --lock: only allow single process running

Examples:
::
    cd code/postconvertor && TARS_ENV=www2 /opt/venv/bin/python -m kipp.runner -t 30 -l "/opt/venv/bin/python PostConvertor.py"
"""

from __future__ import unicode_literals

import argparse
import subprocess
import sys
import inspect
import time
import os
import signal
import traceback
from datetime import timedelta
from random import randint
from textwrap import dedent

from kipp3.options import options as opt

opt.patch_utilities()

from kipp3.aio import Event, run_until_complete
from kipp3.libs.aio import KippAIOTimeoutError
from kipp3.utils import (
    ThreadPoolExecutor,
    check_is_allow_to_running,
    generate_validate_fname,
    get_logger,
    EmailSender,
    utcnow,
)

from .exceptions import (
    KippRunnerTimeoutException,
    KippRunnerException,
    KippRunnerSIGTERMException,
)
from .models import RunStatsMonitor


RECEIVERS = ("lcai@movoto.com",)


def is_need_to_clean_old_records():
    return randint(0, 1000) == 1


def clean_monitor_logs():
    dt_range = timedelta(days=-30)
    if is_need_to_clean_old_records():
        opt.runner_monitor.clean_logs_by_timedelta(dt_range)


def _process_runner(process, evt):
    def _set_evt(futu):
        evt.set()

    f = opt.executor.submit(process.communicate)
    f.add_done_callback(_set_evt)
    return f


def kill_process(process):
    if not process:
        return

    try:
        os.killpg(os.getpgid(process.pid), signal.SIGTERM)
    except OSError:
        pass


def send_alert_email(msg):
    get_logger().info("try to send runner alert email...")
    if opt.debug:
        receivers = "lcai@movoto.com"
    else:
        receivers = ",".join(RECEIVERS)

    content = dedent(
        """
        time: {dt}
        command: {command}
        error: {err}
        """.format(
            dt=utcnow().strftime("%Y-%m-%dT%H:%M:%S"), command=opt.command, err=msg
        )
    )

    opt.sender.send_email(
        mail_from="data@movoto.com",
        mail_to=receivers,
        subject="DATA Monitoring: Runner got critical error",
        content=content,
    )


def wait_process_done(process, timeout):
    evt = Event()
    f_p = _process_runner(process, evt)
    try:
        futu = evt.wait(timeout=opt.timeout)
        run_until_complete(futu)
        futu.result()
    except KippAIOTimeoutError:
        try:
            kill_process(process)
        except OSError:
            return f_p.result()
        else:
            raise KippRunnerTimeoutException(
                "process exceeds timeout {}s".format(timeout)
            )
    else:
        return f_p.result()
    finally:
        opt.executor.shutdown()


def handle_signal_quit(signal, frame):
    err_msg = "quit by signal {}:\n{}".format(signal, inspect.getframeinfo(frame))
    if signal:
        raise KippRunnerSIGTERMException(err_msg)


def catch_sys_quit_signal():
    signal.signal(signal.SIGTERM, handle_signal_quit)


def runner(command):
    process = err_msg = None
    opt.set_option("runner_command_start_at", time.time())
    try:
        opt.set_option(
            "runner_monitor", RunStatsMonitor(command=command, args=sys.argv[1:-1])
        )
        clean_monitor_logs()
        catch_sys_quit_signal()
        opt.runner_monitor.start()
        get_logger().info("kipp.runner for %s", command)
        opt.set_option(
            "runner_command_start_at", time.time()
        )  # override start_at before real process starting
        process = subprocess.Popen(
            [command],
            shell=True,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            preexec_fn=os.setsid,
        )
        if opt.timeout:
            r = wait_process_done(process, opt.timeout)
        else:
            r = process.communicate()

        if process.returncode != 0:
            err_msg = r[1]
            raise RuntimeError(err_msg)
    except BaseException as err:
        get_logger().exception(err)
        err_msg = traceback.format_exc()
        opt.runner_monitor.fail(err_msg)
        # Not send email in kipp.runner
        # if time.time() - opt.runner_command_start_at < opt.minimal_running_seconds:
        #     send_alert_email(err_msg)

        raise
    else:
        get_logger().info("successed: %s", command)
        opt.runner_monitor.success()
    finally:
        kill_process(process)


def setup_settings():
    opt.set_option("executor", ThreadPoolExecutor(2))
    opt.set_option("sender", EmailSender(host=opt.SMTP_HOST))


def setup_arguments():
    opt.add_argument("-t", "--timeout", type=int, default=0, help="seconds")
    opt.add_argument(
        "-ms",
        "--minimal_running_seconds",
        type=int,
        default=30,
        help="minimal running seconds",
    )
    opt.add_argument(
        "-l",
        "--lock",
        action="store_true",
        default=False,
        help="only allow single running",
    )
    opt.add_argument("--debug", action="store_true", default=False)
    opt.add_argument("command", nargs=argparse.REMAINDER)
    opt.parse_args()
    if not opt.command:
        raise AttributeError("You should run like ``python -m runner <COMMAND>``")

    opt.set_option("command", " ".join(opt.command))


def main():
    setup_arguments()
    setup_settings()

    if opt.lock:
        lock_fname = generate_validate_fname(opt.command)
        fp = check_is_allow_to_running(lock_fname)
        if not fp:
            raise KippRunnerException("another process is still running")

    runner(opt.command)
